/* This file is part of Spectacle, the KDE screenshot utility
 * SPDX-FileCopyrightText: 2019 Boudhayan Gupta <bgupta@kde.org>
 * SPDX-License-Identifier: LGPL-2.0-or-later
 */

#include "ImagePlatform.h"

ImagePlatform::ImagePlatform(QObject *parent)
    : QObject(parent)
{
}

#include "moc_ImagePlatform.cpp"
